@@ -4,7 +4,7 @@ package WebService::Avalara::AvaTax;
 
 use Modern::Perl '2011';    ## no critic (Modules::ProhibitUseQuotedVersion)
 
-our $VERSION = '0.004';     # TRIAL VERSION
+our $VERSION = '0.005';     # TRIAL VERSION
 use utf8;
 
 #pod =head1 SYNOPSIS
@@ -28,23 +28,187 @@ use utf8;
 #pod =cut
 
 use Const::Fast;
-use Package::Stash;
+use English '-no_match_vars';
+use Log::Report;
+use LWPx::UserAgent::Cached;
 use Moo;
-use Types::Standard qw(Bool HashRef InstanceOf);
+use MooX::Types::MooseLike::Email 'EmailAddressLoose';
+use Package::Stash;
+use Scalar::Util 'blessed';
+use Sys::Hostname;
+use Types::Standard qw(Bool HashRef InstanceOf Str);
+use URI;
+use XML::Compile::SOAP::WSS;
+use XML::Compile::SOAP11;
+use XML::Compile::SOAP12;
+use XML::Compile::WSDL11;
+use XML::Compile::Transport::SOAPHTTP;
 use namespace::clean;
-with 'WebService::Avalara::AvaTax::Role::Connection';
 
-#pod =head1 OTHER METHODS AND ATTRIBUTES
+#pod =method new
 #pod
-#pod Because this class consumes the
-#pod C<WebService::Avalara::AvaTax::Role::Connection> role, it also has
-#pod that role's attributes and methods. Please consult that module for attributes
-#pod important for establishing a connection such as C<username> and C<password>.
-#pod You can also use other attributes, such as the C<user_agent>
-#pod attribute to access the L<LWP::UserAgent|LWP::UserAgent> used to retrieve and
-#pod call the AvaTax service, or the C<wsdl> method to access the underlying
-#pod L<XML::Compile::WSDL11|XML::Compile::WSDL11> object.
+#pod Builds a new AvaTax web service client. See the L</ATTRIBUTES> section for
+#pod description of its named parameters.
 #pod
+#pod =attr username
+#pod
+#pod The Avalara email address used for authentication. Required.
+#pod
+#pod =cut
+
+has username => ( is => 'ro', isa => EmailAddressLoose, required => 1 );
+
+#pod =attr password
+#pod
+#pod The password used for Avalara authentication. Required.
+#pod
+#pod =cut
+
+has password => ( is => 'ro', isa => Str, required => 1 );
+
+#pod =attr endpoint
+#pod
+#pod A L<URI|URI> object indicating the AvaTax WSDL file to load. Defaults to
+#pod L<https://development.avalara.net/tax/taxsvc.wsdl>. For production API access
+#pod one should set this to L<https://avatax.avalara.net/tax/taxsvc.wsdl>.
+#pod
+#pod =cut
+
+has endpoint => (
+    is     => 'lazy',
+    isa    => InstanceOf ['URI'],
+    coerce => sub {
+        defined blessed( $_[0] )
+            && $_[0]->isa('URI') ? $_[0] : URI->new( $_[0] );
+    },
+    default =>
+        sub { URI->new('https://development.avalara.net/tax/taxsvc.wsdl') },
+);
+
+#pod =attr debug
+#pod
+#pod When set to true, the L<Log::Report|Log::Report> dispatcher used by
+#pod L<XML::Compile|XML::Compile> and friends is set to I<DEBUG> mode.
+#pod
+#pod =cut
+
+has debug => (
+    is      => 'ro',
+    isa     => Bool,
+    default => 0,
+    trigger =>
+        sub { dispatcher( mode => ( $_[1] ? 'DEBUG' : 'NORMAL' ), 'ALL' ) },
+);
+
+#pod =attr user_agent
+#pod
+#pod An instance of an L<LWP::UserAgent|LWP::UserAgent> (sub-)class. You can
+#pod use your own subclass to add features such as caching or enhanced logging.
+#pod
+#pod If you do not specify a C<user_agent> then we default to an
+#pod L<LWPx::UserAgent::Cached|LWPx::UserAgent::Cached> with its C<ssl_opts>
+#pod parameter set to C<< {verify_hostname => 0} >>.
+#pod
+#pod =cut
+
+has user_agent => (
+    is      => 'lazy',
+    isa     => InstanceOf ['LWP::UserAgent'],
+    default => sub {
+        LWPx::UserAgent::Cached->new( ssl_opts => { verify_hostname => 0 } );
+    },
+);
+
+#pod =attr wsdl
+#pod
+#pod After construction, you can retrieve the created
+#pod L<XML::Compile::WSDL11|XML::Compile::WSDL11> instance.
+#pod
+#pod Example:
+#pod
+#pod     my $wsdl = $avatax->wsdl;
+#pod     my @soap_operations = map { $_->name } $wsdl->operations;
+#pod
+#pod =cut
+
+has wsdl => (
+    is       => 'lazy',
+    isa      => InstanceOf ['XML::Compile::WSDL11'],
+    init_arg => undef,
+);
+
+sub _build_wsdl {
+    my $self = shift;
+    return XML::Compile::WSDL11->new(
+        $self->user_agent->get( $self->endpoint )->content );
+}
+
+has _soap_service => ( is => 'ro', isa => Str, default => 'TaxSvc' );
+has _soap_port    => ( is => 'ro', isa => Str, default => 'TaxSvcSoap' );
+
+has _wss => (
+    is      => 'ro',
+    isa     => InstanceOf ['XML::Compile::SOAP::WSS'],
+    default => sub { XML::Compile::SOAP::WSS->new },
+);
+
+has _auth => ( is => 'lazy', isa => InstanceOf ['XML::Compile::WSS'] );
+
+sub _build__auth {
+    my $self = shift;
+    my $wss  = $self->_wss;
+    return $wss->basicAuth( map { ( $_ => $self->$_ ) }
+            qw(username password) );
+}
+
+has _transport =>
+    ( is => 'lazy', isa => InstanceOf ['XML::Compile::Transport::SOAPHTTP'] );
+
+sub _build__transport {
+    my $self = shift;
+    my $wss  = $self->_wss;
+    my $wsdl = $self->wsdl;
+    my $auth = $self->_auth;
+
+    my $user_agent  = $self->user_agent;
+    my %soap_params = (
+        service => $self->_soap_service,
+        port    => $self->_soap_port,
+    );
+    my $endpoint_uri = URI->new( $wsdl->endPoint(%soap_params) );
+    $user_agent->add_handler(
+        request_prepare => sub {
+            $_[0]->header( SOAPAction =>
+                    $wsdl->operation( $self->_operation_name, %soap_params )
+                    ->soapAction, );
+        },
+        (   m_method => 'POST',
+            map { ( "m_$_" => $endpoint_uri->$_ ) } qw(scheme host_port path),
+        ),
+    );
+    return XML::Compile::Transport::SOAPHTTP->new(
+        address    => $endpoint_uri,
+        user_agent => $self->user_agent,
+    );
+}
+
+has _operation_name => ( is => 'rw', isa => Str, default => q{} );
+
+sub _call {
+    my ( $self, $operation, @params ) = @_;
+    $self->_operation_name($operation);
+    return $self->wsdl->call(
+        $operation => {
+            Profile => {
+                Client => "$PROGRAM_NAME," . ( $main::VERSION // q{} ),
+                Adapter => __PACKAGE__ . q{,} . ( $VERSION // q{} ),
+                Machine => hostname(),
+            },
+            parameters => {@params},
+        },
+    );
+}
+
 #pod =head1 SEE ALSO
 #pod
 #pod =over
@@ -66,21 +230,17 @@ with 'WebService::Avalara::AvaTax::Role::Connection';
 #pod
 #pod =back
 #pod
-#pod =for Pod::Coverage BUILD
-#pod
 #pod =head1 METHODS
 #pod
-#pod Available method names are dynamically loaded from the AvaTax WSDL file's
-#pod operations, and can be passed either a hash or reference to a hash with the
-#pod necessary parameters. In scalar context they return a reference to a hash
-#pod containing the results of the SOAP call; in list context they return the
-#pod results hashref and an L<XML::Compile::SOAP::Trace|XML::Compile::SOAP::Trace>
+#pod Aside from the L</new> method, available method names are dynamically loaded
+#pod from the AvaTax WSDL file's operations and can be passed either a hash or
+#pod reference to a hash with the necessary parameters. In scalar context they
+#pod return a reference to a hash containing the results of the SOAP call; in list
+#pod context they return the results hashref and an
+#pod L<XML::Compile::SOAP::Trace|XML::Compile::SOAP::Trace>
 #pod object suitable for debugging and exception handling.
 #pod
-#pod As of this writing the following operations are published in the Avalara
-#pod AvaTax schema. Consult
-#pod L<Avalara's developer site|http://developer.avalara.com/>
-#pod for full documentation on input and output parameters for each operation.
+#pod =for Pod::Coverage BUILD
 #pod
 #pod =cut
 
@@ -111,13 +271,13 @@ sub _method_closure {
         if ( not $self->_compiled->{$method} ) {
             $wsdl->compileCall(
                 $method,
-                transport => $self->transport,
-                service   => $self->soap_service,
-                port      => $self->soap_port,
+                transport => $self->_transport,
+                service   => $self->_soap_service,
+                port      => $self->_soap_port,
             );
             $self->_compiled->{$method} = 1;
         }
-        my ( $answer_ref, $trace ) = $self->call( $method => @_ );
+        my ( $answer_ref, $trace ) = $self->_call( $method => @_ );
 
         #pod =pod
         #pod
@@ -149,7 +309,7 @@ WebService::Avalara::AvaTax - Avalara SOAP interface as compiled Perl methods
 
 =head1 VERSION
 
-version 0.004
+version 0.005
 
 =head1 SYNOPSIS
 
@@ -171,19 +331,18 @@ parameters but will use the same compiled code.
 
 =head1 METHODS
 
-Available method names are dynamically loaded from the AvaTax WSDL file's
-operations, and can be passed either a hash or reference to a hash with the
-necessary parameters. In scalar context they return a reference to a hash
-containing the results of the SOAP call; in list context they return the
-results hashref and an L<XML::Compile::SOAP::Trace|XML::Compile::SOAP::Trace>
+Aside from the L</new> method, available method names are dynamically loaded
+from the AvaTax WSDL file's operations and can be passed either a hash or
+reference to a hash with the necessary parameters. In scalar context they
+return a reference to a hash containing the results of the SOAP call; in list
+context they return the results hashref and an
+L<XML::Compile::SOAP::Trace|XML::Compile::SOAP::Trace>
 object suitable for debugging and exception handling.
 
-As of this writing the following operations are published in the Avalara
-AvaTax schema. Consult
-L<Avalara's developer site|http://developer.avalara.com/>
-for full documentation on input and output parameters for each operation.
+=head2 new
 
-If there is no result then an exception will be thrown.
+Builds a new AvaTax web service client. See the L</ATTRIBUTES> section for
+description of its named parameters.
 
 =head2 get_tax
 
@@ -436,16 +595,45 @@ Example:
         },
     );
 
-=head1 OTHER METHODS AND ATTRIBUTES
+=head1 ATTRIBUTES
 
-Because this class consumes the
-C<WebService::Avalara::AvaTax::Role::Connection> role, it also has
-that role's attributes and methods. Please consult that module for attributes
-important for establishing a connection such as C<username> and C<password>.
-You can also use other attributes, such as the C<user_agent>
-attribute to access the L<LWP::UserAgent|LWP::UserAgent> used to retrieve and
-call the AvaTax service, or the C<wsdl> method to access the underlying
-L<XML::Compile::WSDL11|XML::Compile::WSDL11> object.
+=head2 username
+
+The Avalara email address used for authentication. Required.
+
+=head2 password
+
+The password used for Avalara authentication. Required.
+
+=head2 endpoint
+
+A L<URI|URI> object indicating the AvaTax WSDL file to load. Defaults to
+L<https://development.avalara.net/tax/taxsvc.wsdl>. For production API access
+one should set this to L<https://avatax.avalara.net/tax/taxsvc.wsdl>.
+
+=head2 debug
+
+When set to true, the L<Log::Report|Log::Report> dispatcher used by
+L<XML::Compile|XML::Compile> and friends is set to I<DEBUG> mode.
+
+=head2 user_agent
+
+An instance of an L<LWP::UserAgent|LWP::UserAgent> (sub-)class. You can
+use your own subclass to add features such as caching or enhanced logging.
+
+If you do not specify a C<user_agent> then we default to an
+L<LWPx::UserAgent::Cached|LWPx::UserAgent::Cached> with its C<ssl_opts>
+parameter set to C<< {verify_hostname => 0} >>.
+
+=head2 wsdl
+
+After construction, you can retrieve the created
+L<XML::Compile::WSDL11|XML::Compile::WSDL11> instance.
+
+Example:
+
+    my $wsdl = $avatax->wsdl;
+    my @soap_operations = map { $_->name } $wsdl->operations;
 
 =head1 SEE ALSO
 
@@ -470,6 +658,8 @@ order to debug or extend this module.
 
 =for Pod::Coverage BUILD
 
+If there is no result then an exception will be thrown.
+
 =head1 SUPPORT
 
 =head2 Perldoc
